@@ -47,6 +47,7 @@
 /* Hmm. Should there be a netinet pointer to these??? */
 #include <linux/if_ether.h>
 
+#if 0  /* Darn.  This seems to have been removed in later versions */
 #include <linux/version.h>
 /* This only exists in kernels >= 1.3.75 */
 #if LINUX_VERSION_CODE >= 66379
@@ -57,6 +58,12 @@
 #include <linux/socket.h>
 #define SOCKADDR sockaddr
 #endif
+#endif
+/* assume this for later kenels */
+#define HAS_SOCKADDR_PKT
+#include <linux/if_packet.h>
+#define SOCKADDR sockaddr_pkt
+
 
 #include "config.h"
 #include "fsm.h"
