@@ -1,7 +1,7 @@
 /*
  * firewall.c - Packet filtering for diald.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
@@ -146,12 +146,8 @@ static unsigned int slot_end_timeout(FW_Timeslot *slot, time_t *clock)
     	return maxtime;
 }
 
-/*
- * Add a connection to the queue.
- */
-
-
-static void add_connection(FW_unit *unit, FW_ID *id, unsigned int timeout)
+/* Find a connection in the queue */
+static FW_Connection *find_connection(FW_unit *unit, FW_ID *id)
 {
     FW_Connection *c = unit->connections->next;
 
@@ -162,7 +158,19 @@ static void add_connection(FW_unit *unit, FW_ID *id, unsigned int timeout)
 	   break;
 	c = c->next;
     }
-    if (c == unit->connections) {
+    if (c == unit->connections) return 0;
+    else return c;
+}
+
+/*
+ * Add/update a connection in the queue.
+ */
+
+static void add_connection(
+    FW_unit *unit, FW_Connection *c, FW_ID *id, unsigned int timeout, TCP_STATE lflags)
+{
+    /* look for a connection that matches this one */
+    if (c == NULL) {
 	if (timeout > 0) {
 	    /* no matching connection, add one */
 	    c = malloc(sizeof(FW_Connection));
@@ -172,6 +180,8 @@ static void add_connection(FW_unit *unit, FW_ID *id, unsigned int timeout)
 	    }
 	    c->id = *id;
 	    init_timer(&c->timer);
+            c->tcp_state = lflags;
+	    c->unit = unit;
 	    c->timer.data = (int)c;
 	    c->timer.function = (void *)(int)del_connection;
 	    c->next = unit->connections->next;
@@ -181,20 +191,22 @@ static void add_connection(FW_unit *unit, FW_ID *id, unsigned int timeout)
 	    c->timer.expires = timeout;
 	    add_timer(&c->timer);
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-    		syslog(LOG_INFO,"Adding connection %d @ %d - timeout %d",(int)c,
+    		syslog(LOG_INFO,"Adding connection %d @ %ld - timeout %d",(int)c,
 			time(0),timeout);
 	}
     } else {
 	/* found a matching connection, toss it's old timer */
-	del_timer(&c->timer);
 	if (timeout > 0) {
+	    /* reanimating a ghost? */
+	    del_timer(&c->timer);
 	    c->timer.expires = timeout;
 	    add_timer(&c->timer);
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-    		syslog(LOG_INFO,"Adding connection %d @ %d - timeout %d",(int)c,
+    		syslog(LOG_INFO,"Adding connection %d @ %ld - timeout %d",(int)c,
 			time(0),timeout);
 	} else {
 	    /* timeout = 0, so toss the connection */
+	    del_timer(&c->timer);
 	    del_connection(c);
 	}
     }
@@ -207,9 +219,11 @@ static void add_connection(FW_unit *unit, FW_ID *id, unsigned int timeout)
 static void del_connection(FW_Connection *c)
 {
     if (debug&DEBUG_CONNECTION_QUEUE)
-	syslog(LOG_INFO,"Deleting connection %d @ %d",(int)c,time(0));
+	syslog(LOG_INFO,"Deleting connection %d @ %ld",(int)c,time(0));
+
     c->next->prev = c->prev;
     c->prev->next = c->next;
+    free(c);
 }
 
 static void del_impulse(FW_unit *unit)
@@ -222,7 +236,7 @@ static void del_impulse(FW_unit *unit)
 	    unit->impulse.function = (void *)(int)del_impulse;
 	    unit->impulse.expires = impulse_time;
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %d @ %d",unit->impulse_mode,unit->impulse.expires,time(0));
+		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
 	    add_timer(&unit->impulse);
 	}
     } else {
@@ -233,7 +247,7 @@ static void del_impulse(FW_unit *unit)
 	    unit->impulse.function = (void *)(int)del_impulse;
 	    unit->impulse.expires = impulse_fuzz;
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %d @ %d",unit->impulse_mode,unit->impulse.expires,time(0));
+		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
 	    add_timer(&unit->impulse);
 	}
     }
@@ -346,7 +360,7 @@ next_rule: /* try the next filter */
     	unit->impulse.expires = (force)?ftimeout:itimeout;
     	add_timer(&unit->impulse);
 	if (debug&DEBUG_CONNECTION_QUEUE)
-	    syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %d @ %d",unit->impulse_mode,unit->impulse.expires,time(0));
+	    syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
     }
 }
 
@@ -368,11 +382,11 @@ static void log_packet(int accept, struct iphdr *pkt, int len,  int rule)
 
     if (pkt->protocol == IPPROTO_TCP) {
 	syslog(LOG_INFO,
-	    "filter %s rule %d proto %d len %d seq %x ack %x flags %s%s%s%s%s%s packet %s,%d => %s,%d",
+	    "filter %s rule %d proto %d len %d seq %lx ack %lx flags %s%s%s%s%s%s packet %s,%d => %s,%d",
 	    (accept)?"accepted":"ignored",rule,
 	    pkt->protocol,
-	    htons(pkt->tot_len),
-	    htonl(tcp->seq), htonl(tcp->ack_seq),
+	    ntohs(pkt->tot_len),
+	    ntohl(tcp->seq), ntohl(tcp->ack_seq),
 	    (tcp->fin) ? " FIN" : "",
 	    (tcp->syn) ? " SYN" : "",
 	    (tcp->rst) ? " RST" : "",
@@ -390,6 +404,34 @@ static void log_packet(int accept, struct iphdr *pkt, int len,  int rule)
     }
 }
 
+/* Check if we need to reorder IP addresses for cannonical ordering */
+static int ip_direction(struct iphdr *pkt)
+{
+    struct udphdr *udp = (struct udphdr *)((char *)pkt + 4*pkt->ihl);
+    if (ntohl(pkt->saddr) > ntohl(pkt->daddr)
+    || (ntohl(pkt->saddr) == ntohl(pkt->daddr) &&
+       (pkt->protocol == IPPROTO_TCP || pkt->protocol == IPPROTO_UDP)
+       && ntohs(udp->source) > ntohs(udp->dest)))
+	return 2;
+    else
+	return 1;
+}
+
+static void ip_swap_addrs(struct iphdr *pkt)
+{
+    struct udphdr *udp = (struct udphdr *)((char *)pkt + 4*pkt->ihl);
+    unsigned long taddr;
+    unsigned short tport;
+    if (pkt->protocol == IPPROTO_TCP || pkt->protocol == IPPROTO_UDP) {
+	tport = udp->source;
+	udp->source = udp->dest;
+	udp->dest = tport;
+    } 
+    taddr = pkt->saddr;
+    pkt->saddr = pkt->daddr;
+    pkt->daddr = taddr;
+}
+
 void print_filter(FW_Filter *filter)
 {
     int i;
@@ -411,10 +453,15 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
     FW_unit *unit;
     FW_Filters *fw;
     unsigned char *data;
-    FW_ProtocolRule *prule;
+    FW_ProtocolRule *prule, *pprule = 0;
     FW_Term *term;
     int i,v,rule;
+    int direction,opdir;
+    TCP_STATE lflags;
     clock_t clock = time(0);
+    FW_ID id;
+    FW_Connection *conn;
+    struct iphdr * ip_pkt = (struct iphdr *)pkt;
 
     if (!initialized) init_units();
 
@@ -428,6 +475,64 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 
     data = pkt + 4*((struct iphdr *)pkt)->ihl;
 
+    /* Find the correct protocol rule */
+    for (i = 0; i < unit->nrules; i++) {
+	pprule = &unit->prules[i];
+	if (FW_PROTO_ALL(pprule->protocol)
+	|| pprule->protocol == ip_pkt->protocol)
+	break;
+    }
+
+    if (pprule == 0) return -1;	/* No protocol rules? */
+
+    /* Build the connection ID, and set the direction flag */
+    direction = ip_direction(ip_pkt);
+    if (direction == 2) ip_swap_addrs(ip_pkt);
+    for (i = 0; i < FW_ID_LEN; i++)
+	id.id[i] = (FW_IN_DATA(pprule->codes[i])?data:pkt)
+		    [FW_OFFSET(pprule->codes[i])];
+    if (direction == 2) ip_swap_addrs(ip_pkt);
+    conn = find_connection(unit,&id);
+    opdir = (direction==1)?2:1;
+
+    /* Do the TCP liveness changes */
+    if (ip_pkt->protocol == IPPROTO_TCP) {
+	struct tcphdr *tcp = (struct tcphdr *)((char *)ip_pkt + 4*ip_pkt->ihl);
+	int tcp_data_len = len - (4*ip_pkt->ihl + sizeof(struct tcphdr));
+
+	if (conn) {
+	    lflags = conn->tcp_state;
+	} else {
+	    lflags.fin_seq[0] = lflags.fin_seq[1] = 0;
+	    lflags.tcp_flags = 0;
+            lflags.saw_fin = 0;
+	}
+	if (tcp->rst) {
+	    lflags.saw_fin = 0;
+	    lflags.tcp_flags = 0;
+	} else if (tcp->fin) {
+	    lflags.saw_fin |= direction;
+	    lflags.fin_seq[direction-1] = ntohl(tcp->seq)+tcp_data_len+1;
+	} else if (tcp->syn || tcp_data_len > 0) {
+	    /* Either we have a SYN packet, or we have a data carrying
+   	     * packet. In either case we want to declare this direction live.
+	     * FIXME. It is possible that this should set saw_fin to 0 in the
+	     * case that we have a SYN packet.
+	     */
+	    lflags.saw_fin &= ~direction;
+	    lflags.tcp_flags |= direction;
+	}
+	if ((lflags.saw_fin & opdir) && (tcp->ack)) {
+	    if (lflags.fin_seq[opdir-1] == ntohl(tcp->ack_seq)) {
+		lflags.tcp_flags &= ~direction;
+	    }
+	}
+
+	if (conn) {
+	    conn->tcp_state = lflags;
+	}
+    }
+
     rule = 1;
     while (fw) {
 #ifdef 0
@@ -437,13 +542,15 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 	if ((unit->up && fw->filt.type == FW_TYPE_BRINGUP)
 	   || (!unit->up && fw->filt.type == FW_TYPE_KEEPUP)
 	   || !in_slot(fw->filt.times,&clock)
-	   || fw->filt.type == FW_TYPE_IMPULSE)
+	   || fw->filt.type == FW_TYPE_IMPULSE
+	   || fw->filt.type == FW_TYPE_UP
+	   || fw->filt.type == FW_TYPE_DOWN)
 	    goto next_rule;
 
 	/* Check the protocol rule */
 	prule = &unit->prules[fw->filt.prule];
 	if (!(FW_PROTO_ALL(prule->protocol)
-	|| prule->protocol == ((struct iphdr *)pkt)->protocol))
+	|| prule->protocol == ip_pkt->protocol))
 	    goto next_rule;
 
 	/* Check the terms */
@@ -454,7 +561,10 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 		if (fw == NULL) break;
 	    }
 	    term = &fw->filt.terms[i];
-	    v = (ntohl(*(int *)(&(FW_IN_DATA(term->offset)?data:pkt)
+	    if (FW_TCP_STATE(term->offset))
+		v = (lflags.tcp_flags >> term->shift) && term->mask;
+	    else
+	        v = (ntohl(*(int *)(&(FW_IN_DATA(term->offset)?data:pkt)
 				  [FW_OFFSET(term->offset)]))
 		    >> term->shift) & term->mask;
 #ifdef 0
@@ -471,8 +581,8 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 	    switch (term->op) {
 	    case FW_EQ: if (v != term->test) goto next_rule; break;
 	    case FW_NE: if (v == term->test) goto next_rule; break;
-	    case FW_GE: if (v >= term->test) goto next_rule; break;
-	    case FW_LE: if (v <= term->test) goto next_rule; break;
+	    case FW_GE: if (v < term->test) goto next_rule; break;
+	    case FW_LE: if (v > term->test) goto next_rule; break;
 	    }
 	}
 	/* Ok, we matched a rule. What are we suppose to do? */
@@ -480,16 +590,11 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 	if (fw->filt.log)
 #endif
         if (debug&DEBUG_FILTER_MATCH)
-	    log_packet(fw->filt.type!=FW_TYPE_IGNORE,(struct iphdr *)pkt,len,rule);
+	    log_packet(fw->filt.type!=FW_TYPE_IGNORE,ip_pkt,len,rule);
 
 	/* Check if this entry goes into the queue or not */
 	if (fw->filt.type != FW_TYPE_IGNORE) {
-	    /* Store the connection. Timeout = 0 means delete. */
-	    FW_ID id;
-	    for (i = 0; i < FW_ID_LEN; i++)
-		id.id[i] = (FW_IN_DATA(prule->codes[i])?data:pkt)
-			    [FW_OFFSET(prule->codes[i])];
-	    add_connection(unit,&id,fw->filt.timeout);
+	    add_connection(unit,conn,&id,fw->filt.timeout,lflags);
 	}
 	/* Return 1 if accepting rule with non zero timeout, 0 otherwise */
 	return (fw->filt.type != FW_TYPE_IGNORE && fw->filt.timeout > 0);
@@ -500,13 +605,15 @@ next_rule: /* try the next filter */
     }
     /* Failed to match any rule. This means we ignore the packet */
     if (debug&DEBUG_FILTER_MATCH)
-        log_packet(0,(struct iphdr *)pkt,len,0);
+        log_packet(0,ip_pkt,len,0);
     return 1;
 }
 
 static void pcountdown(long secs)
 {
-    fprintf(mon_fp,"%02ld:%02ld:%02ld\n",secs/3600,(secs/60)%60,secs%60);
+    char buf[10];
+    sprintf(buf,"%02ld:%02ld:%02ld\n",secs/3600,(secs/60)%60,secs%60);
+    if (monitors) mon_write(buf,9);
 }
 
 int ctl_firewall(int op, struct firewall_req *req)
@@ -525,10 +632,13 @@ int ctl_firewall(int op, struct firewall_req *req)
     case IP_FW_QFLUSH:
 	if (!req) return -1; /* ERRNO */
 	{
-	    FW_Connection *c;
-	    while ((c = unit->connections->next) != unit->connections) {
+	    FW_Connection *c,*cn;
+	    for (c = unit->connections->next;
+	    c != unit->connections; c = cn) {
+		cn = c->next;
 		del_timer(&c->timer);
-		del_connection(c);
+		c->next->prev = c->prev;
+		c->prev->next = c->next;
 		free((void *)c);
 	    }
 	    return 0;
@@ -543,7 +653,7 @@ int ctl_firewall(int op, struct firewall_req *req)
 		|| (unit->force == 0
 		    && !(unit->up && unit->impulse_mode == 0
 		   	 && (impulse_init_time > 0 || impulse_time > 0))
-		    && unit->connections->next == unit->connections));
+		    && unit->connections == unit->connections->next));
 
 
     case IP_FW_PFLUSH:
@@ -591,11 +701,13 @@ int ctl_firewall(int op, struct firewall_req *req)
 	    FW_Connection *c;
 	    char saddr[20], daddr[20];
     	    struct in_addr addr;
-	    syslog(LOG_INFO,"up = %d, forcing = %d, impulse = %d, iitime = %d, itime = %d, ifuzz = %d, itimeout = %d, timeout = %d, next alarm = %d",
+	    syslog(LOG_INFO,"up = %d, forcing = %d, impulse = %d, iitime = %d, itime = %d, ifuzz = %d, itimeout = %ld, timeout = %ld, next alarm = %d",
 		unit->up,unit->force, unit->impulse_mode, impulse_init_time, impulse_time,
 		impulse_fuzz,
-		unit->impulse.expected-tstamp,unit->force_etime-atime,next_alarm());
+		unit->impulse.expected-tstamp,unit->force_etime-atime,
+		next_alarm());
 	    for (c=unit->connections->next; c!=unit->connections; c=c->next) {
+		if (c->timer.next == 0) c->timer.expected = tstamp;
                 addr.s_addr = c->id.id[1] + (c->id.id[2]<<8)
                         + (c->id.id[3]<<16) + (c->id.id[4]<<24);
                 strcpy(saddr,inet_ntoa(addr));
@@ -603,16 +715,20 @@ int ctl_firewall(int op, struct firewall_req *req)
                         + (c->id.id[7]<<16) + (c->id.id[8]<<24);
                 strcpy(daddr,inet_ntoa(addr));
                 syslog(LOG_INFO,
-                        "ttl %d, %d - %s/%d => %s/%d",
+                        "ttl %ld, %d - %s/%d => %s/%d (tcp state ([%lx,%lx] %d,%d))",
                         c->timer.expected-tstamp, c->id.id[0],
                         saddr, c->id.id[10]+(c->id.id[9]<<8),
-                        daddr, c->id.id[12]+(c->id.id[11]<<8));
+                        daddr, c->id.id[12]+(c->id.id[11]<<8),
+			c->tcp_state.fin_seq[0],
+			c->tcp_state.fin_seq[1],
+			c->tcp_state.saw_fin,
+			c->tcp_state.tcp_flags);
 	    }
 	    return 0;
 	}
 	return 0;
     case IP_FW_MCONN:
-	if (!req || !mon_fp) return -1; /* ERRNO */
+	if (!req || !monitors) return -1; /* ERRNO */
 	{
 	    unsigned long atime = time(0);
             unsigned long tstamp = timestamp();
@@ -622,16 +738,20 @@ int ctl_firewall(int op, struct firewall_req *req)
 	    char proto[20];
     	    struct protoent *pent;
     	    struct in_addr addr;
+            char buf[1024];
 
-	    fprintf(mon_fp,"STATUS\n%d\n%d\n%d\n%d\n%d\n%d\n",
+	    sprintf(buf,"STATUS\n%d\n%d\n%d\n%d\n%d\n%d\n",
 		unit->up, unit->force, unit->impulse_mode,
 		impulse_init_time, impulse_time,
 		impulse_fuzz);
+	    if (monitors) mon_write(buf,strlen(buf));
+
 	    pcountdown((unit->impulse_mode)?(unit->impulse.expected-tstamp):0);
 	    pcountdown(unit->force_etime-atime);
 	    pcountdown(next_alarm());
-	    fprintf(mon_fp,"QUEUE\n");
+	    if (monitors) mon_write("QUEUE\n",6);
 	    for (c=unit->connections->next; c!=unit->connections; c=c->next) {
+		if (c->timer.next == 0) continue;
                 addr.s_addr = c->id.id[1] + (c->id.id[2]<<8)
                         + (c->id.id[3]<<16) + (c->id.id[4]<<24);
                 strcpy(saddr,inet_ntoa(addr));
@@ -643,14 +763,14 @@ int ctl_firewall(int op, struct firewall_req *req)
 		else sprintf(proto,"%d",c->id.id[0]);
 		sport = c->id.id[10]+(c->id.id[9]<<8);
 		dport = c->id.id[12]+(c->id.id[11]<<8);
-                fprintf(mon_fp,
+                sprintf(buf,
                         "%-4s  %15s/%-5d  %15s/%-5d  ",
                         proto, saddr, sport,
                         daddr, dport);
+                if (monitors) mon_write(buf,strlen(buf));
 		pcountdown(c->timer.expected-tstamp);
 	    }
-	    fprintf(mon_fp,"END QUEUE\n");
-	    fflush(mon_fp);
+	    if (monitors) mon_write("END QUEUE\n",10);
 	    return 0;
 	}
 	return 0;
@@ -690,11 +810,13 @@ int ctl_firewall(int op, struct firewall_req *req)
 	}
     case IP_FW_UP:
 	unit->up = 1;
+	run_ip_up();
 	fw_force_update(unit);
 	fw_impulse_update(unit,1);
 	return 0;
 
     case IP_FW_DOWN:
+	if (unit->up) run_ip_down();
 	unit->up = 0;
 	/* turn off the impulse generator */
 	del_timer(&unit->impulse);
