@@ -1,7 +1,7 @@
 /*
  * options.c - Option parsing code.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
@@ -44,10 +44,15 @@ int default_route = 0;
 int dynamic_addrs = 0;
 int dynamic_mode = DMODE_REMOTE_LOCAL;
 int rotate_devices = 0;
+int two_way = 0;
+int give_way = 0;
+int do_reroute = UNSAFE_ROUTING;
+int proxyarp = 0;
 
 int connect_timeout = 60;
 int disconnect_timeout = 60;
 int redial_timeout = DEFAULT_DIAL_DELAY;
+int nodev_retry_timeout = 1;
 int stop_dial_timeout = 60;
 int kill_timeout = 60;
 int start_pppd_timeout = 60;
@@ -55,6 +60,9 @@ int stop_pppd_timeout = 60;
 int first_packet_timeout = DEFAULT_FIRST_PACKET_TIMEOUT;
 int retry_count = 0;
 int died_retry_count = 1;
+int redial_backoff_start = -1;
+int redial_backoff_limit = 600;
+int dial_fail_limit = 0;
 
 
 struct {
@@ -83,9 +91,15 @@ struct {
     {"defaultroute","",0,&default_route,set_flag},
     {"addroute","<script>",1,&addroute,set_str},
     {"delroute","<script>",1,&delroute,set_str},
+    {"proxyarp","",0,&proxyarp,set_flag},
     {"ip-up","<script>",1,&ip_up,set_str},
     {"ip-down","<script>",1,&ip_down,set_str},
+    {"two-way","",0,&two_way,set_flag},
+    {"give-way","",0,&give_way,set_flag},
+    {"reroute","",0,&do_reroute,set_flag},
+    {"-reroute","",0,&do_reroute,clear_flag},
 /* Modem options */
+    {"device","<device>",1,0,add_device},
     {"connect","<script>",1,&connector,set_str},
     {"disconnect","<script>",1,&disconnector,set_str},
     {"lock","",0,&lock_dev,set_flag},
@@ -100,6 +114,7 @@ struct {
     {"connect-timeout","<timeout>",1,&connect_timeout,set_int},
     {"disconnect-timeout","<timeout>",1,&disconnect_timeout,set_int},
     {"redial-timeout","<timeout>",1,&redial_timeout,set_int},
+    {"nodev-retry-timeout","<timeout>",1,&nodev_retry_timeout,set_int},
     {"stop-dial-timeout","<timeout>",1,&stop_dial_timeout,set_int},
     {"kill-timeout","<timeout>",1,&kill_timeout,set_int},
     {"start-pppd-timeout","<timeout>",1,&start_pppd_timeout,set_int},
@@ -107,6 +122,9 @@ struct {
     {"first-packet-timeout","<timeout>",1,&first_packet_timeout,set_int},
     {"retry-count","<count>",1,&retry_count,set_int},
     {"died-retry-count","<count>",1,&died_retry_count,set_int},
+    {"redial-backoff-start","<count>",1,&redial_backoff_start,set_int},
+    {"redial-backoff-limit","<time>",1,&redial_backoff_limit,set_int},
+    {"dial-fail-limit","<count>",1,&dial_fail_limit,set_int},
     {"prule","<name> <protocol> <spec>",3,0,&parse_prule},
     {"var","<name> <spec>",2,0,&parse_var},
     {"restrict","<start-time> <end-time> <weekday> <day> <month>",5,0,&parse_restrict},
@@ -124,6 +142,58 @@ struct {
     {0,0,0,0}
 };
 
+void init_vars()
+{
+    devices = 0;
+    device_count = 0;
+    inspeed = DEFAULT_SPEED;
+    mtu = DEFAULT_MTU;
+    connector = 0;
+    disconnector = 0;
+    local_ip = 0;
+    local_addr = 0;
+    orig_local_ip = 0;
+    netmask = 0;
+    remote_ip = 0;
+    orig_remote_ip = 0;
+    addroute = 0;
+    delroute = 0;
+    ip_up = 0;
+    ip_down = 0;
+    acctlog = 0;
+    pidlog = "diald.pid";
+    fifoname = 0;
+    acctfp = 0;
+    mode = MODE_SLIP;
+    debug = 0;
+    modem = 0;
+    crtscts = 0;
+    daemon = 1;
+    slip_encap = 0;
+    lock_dev = 0;
+    default_route = 0;
+    dynamic_addrs = 0;
+    dynamic_mode = DMODE_REMOTE_LOCAL;
+    rotate_devices = 0;
+    two_way = 0;
+    do_reroute = UNSAFE_ROUTING;
+    proxyarp = 0;
+    connect_timeout = 60;
+    disconnect_timeout = 60;
+    redial_timeout = DEFAULT_DIAL_DELAY;
+    nodev_retry_timeout = 1;
+    stop_dial_timeout = 60;
+    kill_timeout = 60;
+    start_pppd_timeout = 60;
+    stop_pppd_timeout = 60;
+    first_packet_timeout = DEFAULT_FIRST_PACKET_TIMEOUT;
+    retry_count = 0;
+    died_retry_count = 1;
+    redial_backoff_start = -1;
+    redial_backoff_limit = 600;
+    dial_fail_limit = 0;
+}
+
 void set_int(int *var, char **argv)
 {
     *var = atoi(*argv);
@@ -186,38 +256,35 @@ void read_config_file(int *var, char **argv)
 void usage(void)
 {
     int i;
-    syslog(LOG_ERR,"usage: diald modem-device1 [modem-device2 ...] [options...] [-- [pppd options...]]\n");
+    syslog(LOG_ERR,"usage: diald [modem-device1] [modem-device2 ...] [options...] [-- [pppd options...]]\n");
     syslog(LOG_ERR,"where valid options are:");
     for (i = 0; commands[i].str; i++)
         syslog(LOG_ERR,"    %s %s",commands[i].str,commands[i].uargs);
     exit(1);
 }
 
+void add_device(void *var, char **argv)
+{
+    device_count++;
+    if (device_count == 1)
+    	devices = (char **)malloc(sizeof(char *));
+    else
+	devices = (char **)realloc(devices,sizeof(char *)*device_count);
+
+    devices[device_count-1] = strdup(argv[0]);
+
+}
+
 void parse_args(int argc, char *argv[])
 {
     int i;
     struct stat st;
 
-    /* The first thing MUST be a device specification */
-    if (argc == 0) {
-	syslog(LOG_ERR,
-	    "No device specified. You must have at least one device!");
-	usage();
-    }
-
-    /* Get the first device */
-    devices = argv++;
-    argc--;
-    if (stat(devices[0],&st) < 0 || !S_ISCHR(st.st_mode)) {
-	syslog(LOG_ERR,
-	    "Specified device '%s' not a character device.",devices[0]);
-	usage();
-    }
-
-    /* Get any extra devices on the list */
-    for (i = 1; argc > 0 && stat(*argv,&st) == 0 && S_ISCHR(st.st_mode); i++)
+    /* Get any devices from the initial arguments */
+    for (i = 0; argc > 0 && stat(*argv,&st) == 0 && S_ISCHR(st.st_mode); i++) {
+	add_device(0,argv);
 	argc--, argv++;
-    device_count = i;
+    }
 
     /*
      * Ok, parse the options list now,
@@ -261,29 +328,73 @@ void parse_options_file(char *file)
     int argc, i;
     FILE *fp = fopen(file,"r");
 
-    if (fp == NULL) return;	/* no options file */
+    if (fp == NULL) {
+	syslog(LOG_ERR,"Unable to open options file %s: %m",file);
+	return;	/* no options file */
+    }
     while (fscanf(fp,"%1024[^\n]\n",line) != -1) {
 	argc = 0;
-	for (s = line, argc = 0; *s, argc < MAXARGS; s++) {
+	for (s = line, argc = 0; *s || argc < MAXARGS; s++) {
 	    if (*s == ' ' || *s == '\t')
 		*s = 0;
 	    else if (*s == '#' && argc == 0) /* the line is a comment */
 		goto comment;
-	    else if (*s == '\"') {            /* start of a quoted argument */
+	    else if (*s == '\"' || *s == '\'') {
+		char delim = *s;
+		/* start of a quoted argument */
 		s++;
 		argv[argc] = s;
 		while (*s) {
-		    if (*s == '\"') { *s++ = 0; break; }
-		    if (*s == '\\' && s[1] == '\"') s++;
+		    if (*s == delim) { *s++ = 0; break; }
+		    if (*s == '\\' && s[1] != 0) s++;
 		    s++;
 		}
 		s--;
 		/* go back and fix up "quoted" characters */
 		t1 = t2 = argv[argc++];
 		while (*t1) {
-		   if (*t1 == '\\') t1++;
+		   if (*t1 == '\\') {
+			t1++;
+			/* do a translation */
+			switch (*t1) {
+			case 'a': *t1 = '\a'; break;
+			case 'b': *t1 = '\b'; break;
+			case 'f': *t1 = '\f'; break;
+			case 'n': *t1 = '\n'; break;
+			case 'r': *t1 = '\r'; break;
+			case 's': *t1 = ' '; break;
+			case 't': *t1 = '\t'; break;
+			default:
+			    if (*t1 >= '0' && *t1 <= '8') {
+				int val = 0, n;
+				for (n = 0;
+				n < 3 && (*t1 >= '0' && *t1 <= '8');
+				++n) {
+				    val = (val << 3) + (*t1 & 07);
+				    t1++;
+				}
+				*(--t1) = val;
+				break;
+			    }
+			    if (*t1 == 'x') {
+				int val = 0, n, digit;
+				t1++;
+				for (n = 0; n < 2 && isxdigit(*t1); ++n) {
+				    digit = toupper(*t1) -'0';
+				    if (digit > 10)
+					digit += '0' + 10 - 'A';
+				    val = (val << 4) + digit;
+				    t1++;
+				}
+				*(--t1) = val;
+				break;
+			    }
+			    /* character is itself, don't muck with it. */
+			}
+		   }
 		   *t2++ = *t1++;
 		}
+		*t2++ = 0;
 	    } else { /* just a normal word */
 		argv[argc++] = s;
 		while (*s) {
@@ -315,8 +426,21 @@ comment:;
 
 void check_setup()
 {
+    struct stat st;
+    int i;
     int flag = 0;
 
+    if (device_count == 0)
+	flag = 1,
+	syslog(LOG_ERR,
+	    "No device specified. You must have at least one device!");
+
+    for (i = 0; i < device_count; i++)
+	if (stat(devices[i],&st) < 0 || !S_ISCHR(st.st_mode))
+	    flag = 1,
+	    syslog(LOG_ERR,
+		"Specified device '%s' not a character device.",devices[0]);
+
     if (!connector)
 	flag = 1,
 	syslog(LOG_ERR,
